package be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl;

import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Service;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.lowlevel.api.ProtogeniSliceAuthority;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.JFedTrustStore;
import be.iminds.ilabt.jfed.util.RFC3339Util;
import org.slf4j.LoggerFactory;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;

/**
 * ProtoGeniSAUserAndSliceApiWrapper
 */
public class ProtoGeniSAUserAndSliceApiWrapper extends UserAndSliceApiWrapper {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(ProtoGeniSAUserAndSliceApiWrapper.class);
    private List<AnyCredential> userCredentialsNoSubAuth;
    private Map<String, List<AnyCredential>> userCredentialsBySubAuth = new HashMap<String, List<AnyCredential>>();
    private HashMap<String, String> subAuthNamesToUrl = null;

    public ProtoGeniSAUserAndSliceApiWrapper(Logger logger, GeniUserProvider geniUserProvider, JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences) {
        super(logger, geniUserProvider, connectionProvider, jFedPreferences);
    }

    private ProtogeniSliceAuthority sa(Logger logger) {
        return new ProtogeniSliceAuthority(logger, jFedPreferences);
    }

    protected SfaConnection getConnection(Logger logger, String subAuthName) throws JFedException {
        if (subAuthName == null)
            return getConnection(new ApiInfo.Api(ApiInfo.ApiName.PROTOGENI_SA, 1));
        else {

            if (subAuthNamesToUrl == null) {
                if (userCredentialsNoSubAuth == null)
                    throw new NullPointerException("There are no user credentials stored yet, while that is needed to request the names of the sub authorities");

                GeniUrn userUrn;
                if (speaksForUserUrn != null) {
                    userUrn = speaksForUserUrn;
                } else
                    userUrn = geniUserProvider.getLoggedInGeniUser().getUserUrn();

                getSubAuthorityNames(logger, userUrn);
            }

            if (subAuthNamesToUrl == null)
                throw new JFedException("Error in ProtoGeniSAUserAndSliceApiWrapper getSubAuthority: no sub authorities known. subAuthName=" + subAuthName);

            String subAuthUrlStr = subAuthNamesToUrl.get(subAuthName);
            if (subAuthUrlStr == null)
                throw new JFedException("Error in ProtoGeniSAUserAndSliceApiWrapper getSubAuthority: no sub authority for subAuthName=" + subAuthName);

            URL subAuthUrl;
            try {
                subAuthUrl = new URL(subAuthUrlStr);
            } catch (MalformedURLException e) {
                throw new JFedException("Error in ProtoGeniSAUserAndSliceApiWrapper getSubAuthority: no valid URL for subAuthName=" + subAuthName + " subAuthUrlStr=\"" + subAuthUrlStr + "\"", e);
            }


            GeniUser user = geniUserProvider.getLoggedInGeniUser();
//            Server userAuth = user.getUserAuthority();
            JFedTrustStore jFedTrustStore = new JFedTrustStore();
            jFedTrustStore.addTrustedPemCertificateIfNotAdded(user.getUserAuthorityServer().getCertificateChain());

            Service service = ApiInfo.findService(user.getUserAuthorityServer(), ApiInfo.ApiName.PROTOGENI_SA, 1);

            return (SfaConnection) connectionProvider.getConnectionByUrl(jFedTrustStore, user, subAuthUrl,
                    null/*handleUntrustedCallback*/, new ApiInfo.Api(ApiInfo.ApiName.PROTOGENI_SA, 1),
                    service,/*optional authority for debug info*/null);
        }
    }

    public List<AnyCredential> getUserCredentials(Logger logger, GeniUrn userUrn, String subAuthorityName) throws JFedException {
        if (subAuthorityName != null)
            assert userCredentialsNoSubAuth != null;

        SfaConnection con = getConnection(logger, subAuthorityName);
        assert con.getDebugInfo() != null && con.getDebugInfo().getService() != null;

        ProtogeniSliceAuthority.SliceAuthorityReply<AnyCredential> reply;
        if (speaksForCredentials == null || speaksForCredentials.isEmpty()) {
            reply =
                    sa(logger).getCredential(con);
        } else {
            reply =
                    sa(logger).getAnyCredential(con, speaksForCredentials, null, null);

        }

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Could not retrieve user credential from server: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply.getXMLRPCCallDetailsWithCodeValueError(), reply.getGeniResponseCode()
            );

        List<AnyCredential> res = new ArrayList<AnyCredential>();

        res.add(reply.getValue());

        if (speaksForCredentials != null && !speaksForCredentials.isEmpty()) {
            res.addAll(speaksForCredentials);
        }

//        if (speaksForCredential != null) {
//            int i = 0;
//            for (AnyCredential cred : res) {
//                assert cred.validate() : "Credential "+i+" is not valid";
//                i++;
//            }
//        }

        if (subAuthorityName == null)
            this.userCredentialsNoSubAuth = res;
        else
            this.userCredentialsBySubAuth.put(subAuthorityName, res);
        return new ArrayList<AnyCredential>(res);
    }

    public List<AnyCredential> getCachedUserCredentialsForAM() {
        if (userCredentialsNoSubAuth == null)
            return Collections.emptyList();
        else
            return new ArrayList<AnyCredential>(userCredentialsNoSubAuth);
    }

    @Override
    public boolean hasUserCredentials() {
        return hasUserCredentials(userCredentialsNoSubAuth);
    }

    protected List<AnyCredential> getSubAuthCredentials(Logger logger, String subAuthName) throws JFedException {
        List<AnyCredential> res = userCredentialsBySubAuth.get(subAuthName);
        if (res == null) {
            res = getUserCredentials(logger, geniUserProvider.getLoggedInGeniUser().getUserUrn(), subAuthName);
        }
        if (res.isEmpty() || !hasUserCredentials(res)) {
            res = getUserCredentials(logger, geniUserProvider.getLoggedInGeniUser().getUserUrn(), subAuthName);
        }
        return res;
    }

    @Override
    public List<AnyCredential> getUserCredentials(Logger logger, GeniUrn userUrn) throws JFedException {
        return getUserCredentials(logger, userUrn, null/*subAuthorityName*/);
    }

    @Override
    public List<AnyCredential> getSliceCredentials(Logger logger, GeniUrn sliceUrn) throws JFedException {
        List<AnyCredential> userCredentials_forCall;

        if (sliceUrn.getEncodedSubAuthName() == null) {
            userCredentials_forCall = userCredentialsNoSubAuth;
        } else {
            //use subauth credentials for call
            userCredentials_forCall = getSubAuthCredentials(logger, sliceUrn.getEncodedSubAuthName());
        }
        if (userCredentials_forCall == null)
            throw new RuntimeException("Internal jFed error: no correct credential stored. (sliceUrn=" + sliceUrn + ", sliceUrn.getSubAuthName()=" + sliceUrn.getEncodedSubAuthName() + ")");

        ProtogeniSliceAuthority.SliceAuthorityReply<AnyCredential> reply;
        if (speaksForCredentials == null || speaksForCredentials.isEmpty()) {
            reply =
                    sa(logger).getSliceCredential(getConnection(logger, sliceUrn.getEncodedSubAuthName()),
                            AnyCredential.toSingleCredential(userCredentials_forCall),
                            sliceUrn
                    );
        } else {
            reply =
                    sa(logger).getSliceCredential(getConnection(logger, sliceUrn.getEncodedSubAuthName()),
                            userCredentials_forCall,
                            sliceUrn
                    );
        }

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Could not retrieve slice credential from server: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply.getXMLRPCCallDetailsWithCodeValueError(), reply.getGeniResponseCode()
            );

        if (speaksForCredentials == null || speaksForCredentials.isEmpty())
            return Collections.singletonList(reply.getValue());
        else {
            List<AnyCredential> res = new ArrayList<AnyCredential>();
            res.add(reply.getValue());
            res.addAll(speaksForCredentials);
            return res;
        }
    }

    @Override
    public List<GeniUrn> getSlicesForUser(Logger logger, GeniUrn user) throws JFedException {
        List<GeniUrn> res = new ArrayList<GeniUrn>();
        ProtogeniSliceAuthority.SliceAuthorityReply<Hashtable> reply;

        if (!hasUserCredentials()) {
            getUserCredentials(logger, user);
            assert hasUserCredentials();
        }
        if (speaksForCredentials == null || speaksForCredentials.isEmpty()) {
            reply =
                    sa(logger).resolveUser(
                            getConnection(logger, null/*subAuthName*/),
                            AnyCredential.toSingleCredential(userCredentialsNoSubAuth),
                            user);
        } else {
            reply =
                    sa(logger).resolveUser(
                            getConnection(logger, null/*subAuthName*/),
                            userCredentialsNoSubAuth,
                            user);
        }

        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            return res;
        Hashtable userInfo = reply.getValue();
        Vector slicesVect = (Vector) userInfo.get("slices");
        for (Object aSlicesVect : slicesVect) {
            String s = aSlicesVect.toString();
            try {
                res.add(new GeniUrn(s));
            } catch (GeniUrn.GeniUrnParseException e) {
                LOG.error("error in reply: Not a slice urn: \"" + s + "\"   (see ProtoGeniSAUserAndSliceApiWrapper getSlicesForUser)");
                //ignore
            }
        }

        return res;
    }

    @Override
    public List<String> getAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn) throws JFedException {
        List<AnyCredential> userCredentials_forCall;

        if (sliceUrn.getEncodedSubAuthName() == null) {
            //remove sub auth credentials, they are not needed
            userCredentials_forCall = userCredentialsNoSubAuth;
        } else {
            //use subauth credentials for call
            //use non subauth credentials for connection
            userCredentials_forCall = getSubAuthCredentials(logger, sliceUrn.getEncodedSubAuthName());
        }
        if (userCredentials_forCall == null)
            throw new RuntimeException("Internal jFed error: no correct credential stored. (sliceUrn=" + sliceUrn + ", sliceUrn.getSubAuthName()=" + sliceUrn.getEncodedSubAuthName() + ")");

        ProtogeniSliceAuthority.SliceAuthorityReply<Hashtable> reply;

        if (speaksForCredentials == null || speaksForCredentials.isEmpty()) {
            reply =
                    sa(logger).resolveSlice(
                            getConnection(logger, sliceUrn.getEncodedSubAuthName()),
                            AnyCredential.toSingleCredential(userCredentials_forCall),
                            sliceUrn
                    );
        } else {
            reply =
                    sa(logger).resolveSlice(
                            getConnection(logger, sliceUrn.getEncodedSubAuthName()),
                            userCredentials_forCall,
                            sliceUrn
                    );
        }

        if (!reply.getGeniResponseCode().isSuccess()) return null;
        Hashtable val = reply.getValue();
        if (val == null) return null;
        Object comPan = val.get("component_managers");
        if (comPan == null || !(comPan instanceof Vector)) return null;
        @SuppressWarnings("unchecked")
        List<String> res = new ArrayList<String>((Vector) reply.getValue().get("component_managers"));
        return res;
    }

    @Override
    public Date getSliceExpiration(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn) throws JFedException {
        //Resolve of slice does not return expiration
        return null;
    }

    @Override
    public void shareSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, GeniUrn userUrn) throws JFedException {
        if (speaksForCredentials == null || speaksForCredentials.isEmpty()) {
            ProtogeniSliceAuthority.SliceAuthorityReply<Boolean> reply =
                    sa(logger).bindToSlice(
                            getConnection(logger, sliceUrn.getEncodedSubAuthName()),
                            AnyCredential.toSingleCredential(sliceCredentials),
                            userUrn
                    );
        } else {
            ProtogeniSliceAuthority.SliceAuthorityReply<Boolean> reply =
                    sa(logger).bindToSlice(
                            getConnection(logger, sliceUrn.getEncodedSubAuthName()),
                            sliceCredentials,
                            userUrn
                    );
        }
    }

    @Override
    public SliceInfo createSlice(Logger logger, String sliceName, Date expirationDate, String subAuthName) throws JFedException {
        assert sliceName != null;

        List<AnyCredential> userCredentials_forCall;

        if (subAuthName == null) {
            //remove sub auth credentials, they are not needed
            userCredentials_forCall = userCredentialsNoSubAuth;
        } else {
            //use subauth credentials for call
            //use non subauth credentials for connection
            userCredentials_forCall = getSubAuthCredentials(logger, subAuthName);
        }
        if (userCredentials_forCall == null)
            throw new RuntimeException("Internal jFed error: no correct credential stored. (subAuthName=" + subAuthName + ")");

        GeniUrn sliceUrn = GeniUrn.createGeniUrnFromEncodedParts(getLoggedInUserAuthorityServer().getUrnTld(), "slice", sliceName);
        if (subAuthName != null)
            sliceUrn = GeniUrn.createGeniUrnFromEncodedParts(getLoggedInUserAuthorityServer().getUrnTld() + ":" + subAuthName, "slice", sliceName);

        ProtogeniSliceAuthority.SliceAuthorityReply<AnyCredential> reply;
        if (speaksForCredentials == null || speaksForCredentials.isEmpty()) {
            reply = sa(logger).register(
                    getConnection(logger, subAuthName),
                    AnyCredential.toSingleCredential(userCredentials_forCall),
                    sliceUrn);
        } else {
            reply = sa(logger).register(
                    getConnection(logger, subAuthName),
                    userCredentials_forCall,
                    sliceUrn);
        }

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Could not create slice on server: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply
            );

        AnyCredential sliceCredential = reply.getValue();
        boolean mustRenew = expirationDate != null;
        if (expirationDate != null && sliceCredential instanceof SfaCredential) {
            SfaCredential sfaSfaCredential = (SfaCredential) sliceCredential;
            mustRenew =
                    sfaSfaCredential.getExpiresDate() != null && sfaSfaCredential.getExpiresDate().before(expirationDate);
        }

        if (mustRenew) {
            try {
                ProtogeniSliceAuthority.SliceAuthorityReply<AnyCredential> replyRenew =
                        sa(logger).renewSlice(
                                getConnection(logger, subAuthName),
                                sliceCredential,
                                RFC3339Util.dateToRFC3339String(expirationDate, true, true, true));

                if (replyRenew.getGeniResponseCode().isSuccess()) {
                    sliceCredential = replyRenew.getValue(); //credential for new expire time
                } else {
                    LOG.warn("Renew call to set expiration of new Slice date failed. Ignoring. Output: " + replyRenew.getOutput());
                }
            } catch (JFedException e) {
                LOG.error("Ignoring exception when renewing: " + e.getMessage(), e);
            }
        }

        GeniUrn returnedSliceUrn = sliceUrn; //fallback
        if (sliceCredential instanceof SfaCredential) {
            SfaCredential sfaCred = (SfaCredential) sliceCredential;
            returnedSliceUrn = GeniUrn.parse(sfaCred.getTargetUrn());
            if (!returnedSliceUrn.equals(sliceUrn))
                LOG.warn("URN CHANGED: \"Register\" of slice with urn \"" + sliceUrn + "\" returned credential for URN \"" + returnedSliceUrn + "\"");
        }

        List<AnyCredential> sliceCredentials = new ArrayList<AnyCredential>();
        sliceCredentials.add(sliceCredential);
        if (speaksForCredentials != null && !speaksForCredentials.isEmpty())
            sliceCredentials.addAll(speaksForCredentials);
        return new SliceInfo(sliceName, returnedSliceUrn, sliceCredentials);
    }

    @Override
    public List<AnyCredential> renewSlice(Logger logger, List<AnyCredential> sliceCredentials, Date newExpirationDate) throws JFedException {
        //
        GeniUrn sliceUrn = null;
        for (AnyCredential c : sliceCredentials) {
            if (c instanceof SfaCredential) {
                SfaCredential sc = (SfaCredential) c;
                GeniUrn target = GeniUrn.parse(sc.getTargetUrn());
                assert target != null;
                if (sliceUrn == null)
                    sliceUrn = target;
            }
        }
        if (sliceUrn == null)
            LOG.warn("renewSlice: could not find sliceUrn in credential list. cannot determine if this is slice for subAuthority! Assuming no");
        String subAuthName = sliceUrn == null ? null : sliceUrn.getEncodedSubAuthName();

        //assuming that when subauth slice credential is known, URL of subAuth is also known
        if (subAuthName != null)
            assert subAuthNamesToUrl != null;

        ProtogeniSliceAuthority.SliceAuthorityReply<AnyCredential> reply;
        if (speaksForCredentials == null || speaksForCredentials.isEmpty())
            reply =
                    sa(logger).renewSlice(getConnection(logger, subAuthName),
                            AnyCredential.toSingleCredential(sliceCredentials),
                            RFC3339Util.dateToRFC3339String(newExpirationDate, true, true, true));
        else
            reply =
                    sa(logger).renewSlice(getConnection(logger, subAuthName),
                            sliceCredentials,
                            RFC3339Util.dateToRFC3339String(newExpirationDate, true, true, true));

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Could not renew slice on server: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply.getXMLRPCCallDetailsWithCodeValueError(), reply.getGeniResponseCode()
            );

        List<AnyCredential> newSliceCredentials = new ArrayList<AnyCredential>();
        newSliceCredentials.add(reply.getValue());
        if (speaksForCredentials != null && !speaksForCredentials.isEmpty())
            newSliceCredentials.addAll(speaksForCredentials);
        return newSliceCredentials;
    }

    @Override
    public List<String> getSshKeysForUser(Logger logger, GeniUrn userUrn) throws JFedException {
        if (!geniUserProvider.getLoggedInGeniUser().getUserUrn().equals(userUrn))
            throw new JFedException("Protogeni SA API can NOT get keys for other user than logged in user.");

        ProtogeniSliceAuthority.SliceAuthorityReply<List<String>> reply =
                sa(logger).getKeys(getConnection(logger, null/*subAuthName*/), AnyCredential.toSingleCredential(userCredentialsNoSubAuth));
        if (reply.getGeniResponseCode().isSuccess())
            return reply.getValue();
        else
            throw new JFedException("Failed to get SSH keys", reply.getXMLRPCCallDetailsWithCodeValueError(), reply.getGeniResponseCode());
    }

    @Override
    public SubAuthoritySupport getSubAuthoritySupport(Logger logger) throws JFedException {
        return SubAuthoritySupport.SUB_AUTHORITY_OPTIONAL;
    }

    /**
     * Example in ResolveUser reply:
     * "subauthorities" -> {
     * "urn:publicid:IDN+wall2.ilabt.iminds.be:odspracticum+authority+sa" -> "https://www.wall2.ilabt.iminds.be:12369/protogeni/xmlrpc/project/ODSPracticum/sa",
     * "urn:publicid:IDN+wall2.ilabt.iminds.be:fed4fire+authority+sa" -> "https://www.wall2.ilabt.iminds.be:12369/protogeni/xmlrpc/project/Fed4Fire/sa"
     * },
     */
    @Override @SuppressWarnings("unchecked")
    public List<String> getSubAuthorityNames(Logger logger, GeniUrn user) throws JFedException {
        ProtogeniSliceAuthority.SliceAuthorityReply<Hashtable> reply;
        if (speaksForCredentials == null || speaksForCredentials.isEmpty()) {
            reply =
                    sa(logger).resolveUser(
                            getConnection(logger, null/*subAuthName*/),
                            AnyCredential.toSingleCredential(userCredentialsNoSubAuth),
                            user);
        } else {
            reply =
                    sa(logger).resolveUser(
                            getConnection(logger, null/*subAuthName*/),
                            userCredentialsNoSubAuth,
                            user);
        }

        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            return new ArrayList<String>();
        Hashtable userInfo = reply.getValue();

        subAuthNamesToUrl = null;
        if (!userInfo.containsKey("subauthorities")) {
            LOG.warn("ResolveUser did not return \"subauthorities\"");
            return new ArrayList<String>();
        }

        if (userInfo.get("subauthorities") instanceof Hashtable) {
            subAuthNamesToUrl = new HashMap<String, String>();

            Hashtable slicesHt = (Hashtable) userInfo.get("subauthorities");

            for (Map.Entry<String, String> entry :  (Set<Map.Entry<String, String>>) slicesHt.entrySet()) {
                String urnStr = entry.getKey();
                String urlStr = entry.getValue();
                try {
                    GeniUrn urn = new GeniUrn(urnStr);
                    String subAuthName = urn.getEncodedSubAuthName();
                    if (subAuthName != null) {
                        subAuthNamesToUrl.put(subAuthName, urlStr);
                    } else {
                        LOG.warn("error in ResolveUser reply: No subauthority in urn: \"" + urnStr + "\"  (see ProtoGeniSAUserAndSliceApiWrapper getSubAuthorityNames)");
                    }
                } catch (GeniUrn.GeniUrnParseException e) {
                    LOG.warn("error in ResolveUser reply: Not a valid slice urn: \"" + urnStr + "\"   (see ProtoGeniSAUserAndSliceApiWrapper getSubAuthorityNames)");
                    //ignore
                }
            }
        } else {
            LOG.warn("ResolveUser subauthorities is not a Hashtable but a " + userInfo.get("subauthorities").getClass().getName());
        }

        return new ArrayList<String>(subAuthNamesToUrl.keySet());
    }

    @Override
    public List<GeniUrn> getUsersForSubAuthority(Logger logger, String projectName) throws JFedException {
        throw new UnsupportedOperationException("Not supported by API");
    }

    @Override
    public List<GeniUrn> getUsersForSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials) throws JFedException {
        throw new UnsupportedOperationException("Not supported by API");
    }

    @Override
    public void addUsersToSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<GeniUrn> userUrns, String role) throws JFedException {
        throw new RuntimeException("Not implemented");
    }

    @Override
    public boolean isRegisterAggregatesForSliceSupported() {
        return false;
    }
    @Override
    public void registerAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn, GeniUrn aggregateUrn, Collection<GeniUrn> sliverUrns, Date sliverExpirationDate) throws JFedException {
        throw new UnsupportedOperationException("Not supported by API");
    }
    @Override
    public void updateSliverExpirationDateInfoForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn, Collection<GeniUrn> sliverUrns, Date sliverExpirationDate) throws JFedException {
        throw new UnsupportedOperationException("Not supported by API");
    }
    @Override
    public void unregisterAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn, GeniUrn aggregateUrn, Collection<GeniUrn> sliverUrns) throws JFedException {
        throw new UnsupportedOperationException("Not supported by API");
    }
    @Override
    public boolean hasSpeaksForSupport() {
        //the Protogeni API supports speaksFor
        return true;
    }
}
